/**
  OpenGL/GLUT/GLUI "playground" for 2D mesh-based physics.
  
  The multigrid aspect is completely in the pixel shader, and doesn't 
  even show up here.
  
  Dr. Orion Sky Lawlor, olawlor@acm.org, 2009-02-23 (Public Domain)
*/
#include "GL/glew.h" /* GL extension wrangler: http://glew.sourceforge.net/ */
#include "GL/glut.h" /* windows, mice, etc.  http://freeglut.sourceforge.net */
#include <stdlib.h> /* for "rand" */
#include <stdio.h> /* for printf */
#include "osl/vec4.h" /* 3D vector class */
#include "GL/glui.h" /* user interface widgets: http://glui.sourceforge.net/ */

#include "ogl/minicam.h" /* ortho_frame, camera motion, key_down & events */

#ifndef M_PI /* why doesn't windows math.h have this? */
#define M_PI 3.14159265358979
#endif

/* These variables are changed directly by the user interface code below */

/* Enable for anisotropic filtering */
int texaniso=1;

/* Texture pixel format on graphics card */
int texformat=GL_RGBA8; 

/* Texture "wrap mode" in X and Y: */
int wrap_mode=GL_REPEAT;

/* Texture magnification mode/minification mode */
int minfilter_mode=GL_LINEAR_MIPMAP_LINEAR; 
int magfilter_mode=GL_LINEAR; 

/* Return the wall clock time, in seconds */
double oglTimer(void) {return 0.001*glutGet(GLUT_ELAPSED_TIME);}
/* Check if an error has happened--if so, exit */
void oglCheck(const char *where) {
	int err=0;
	if (0!=(err=glGetError())) {
		fprintf(stderr,"OpenGL error %d at %s\n",err,where);
		exit(1);
	}
}

/* Make an OpenGL texture handle, of this size, with this pixel format */
GLuint make_texture(int w,int h,GLenum format=GL_RGBA8)
{
	GLuint tex=0;
	glGenTextures(1,&tex);
	GLenum t=GL_TEXTURE_2D;
	glBindTexture(t,tex);
	char *mem=new char[w*h*4]; memset(mem,0x80,w*h*4);
	glTexImage2D(t,0,format, w,h,0, GL_RGBA,GL_UNSIGNED_BYTE,mem);
	delete[] mem;
	glTexParameteri(t, GL_TEXTURE_MAG_FILTER,GL_LINEAR); 
	glTexParameteri(t, GL_TEXTURE_MIN_FILTER,GL_LINEAR); /* no mipmaps */
	GLenum texWrap=GL_CLAMP_TO_EDGE; /* stop changing at edge */
	glTexParameteri(t, GL_TEXTURE_WRAP_S, texWrap); 
	glTexParameteri(t, GL_TEXTURE_WRAP_T, texWrap);
	glGenerateMipmapEXT(t); /* make mipmaps */
	glBindTexture(t,0); /* back to default texture */
	oglCheck("after make_texture");
	return tex;
}

/* Make a framebuffer object that renders pixels into this texture */
GLuint make_fbo(GLuint tex)
{
	GLuint fb=0;
	glGenFramebuffersEXT(1, &fb);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, 
		GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, tex, 0);
	
	/* Make sure the framebuffer's happy */
	GLenum status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
	if (status!=GL_FRAMEBUFFER_COMPLETE_EXT)
	{ fprintf(stderr,"Error %d in framebuffer creation!\n",status); exit(1); }

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0); /* back to screen render */
	oglCheck("after make_fbo");
	return fb;
}


const unsigned int plasma_size=512;
const unsigned int plasma_mask=plasma_size-1;

/* Tiny little 2D "height image" class */
class terrain {
public:
	/* Size of our height image, in pixels. */
	int w,h;
	
	/* Data in our image: w * h pixels, each one a height. */
	float *ht;
	
	terrain(int w_,int h_) :w(w_), h(h_), ht(new float[w*h]) {}
	
	/* Access the terrain height at pixel (x,y) */
	float &operator() (int x,int y) { return ht[y*w+x];} /* write height */
};

/**
 Random number generator: at the location X,Y (MUST be deterministic!),
 generate a perturbation for a span of "size" pixels.
*/
float perturb(unsigned int X,unsigned int Y, float size) {
	X&=plasma_mask; Y&=plasma_mask; /* makes image wrap properly */
	unsigned int seed=X*2444u + Y*98023u;
	unsigned int seed2=X*24943u + Y*980237u;
	unsigned int val=(seed^(seed2>>8)^(seed>>16)) % 256u;
	float scale=1.0/128; /* fixes magnitude of result */
	
	float falloff=1.0; /* decreases perturbation at small sizes */
	int falloffsize=100;
	if (size<falloffsize) falloff=size*(1.0/falloffsize);

	return 0.3*falloff*scale*(val-128.0);
}

/** Initialize terrain to random value */
void init(terrain &t) {
	for (int y=0;y<t.h;y++)
	for (int x=0;x<t.w;x++)
	{
		t(x,y)=0.5; // initial value to be perturbed
	}
}

/** Recursively draw plasma fractal on this terrain. 
The four corners must already have been set.
       T
   L       R
       B
*/
void plasma(terrain &t,int L,int R, int T, int B) {
	if (R-L<=1) return;
	if (B-T<=1) return; /* no pixels in between to render! */
	
	// Read terrain corner values
	float TL=t(L,T);
	float TR=t(R,T);
	float BL=t(L,B);
	float BR=t(R,B);
	
	// Compute center coordinates
	int X=(L+R)/2, Y=(T+B)/2;
	
	// Compute perturbed centers
	float YL=(TL+BL)*0.5 + perturb(Y,L,B-T);
	float YR=(TR+BR)*0.5 + perturb(Y,R,B-T);
	float TX=(TL+TR)*0.5 + perturb(T,X,R-L);
	float BX=(BL+BR)*0.5 + perturb(B,X,R-L);
	float YX=(TL+TR+BL+BR)*0.25 + perturb(Y,X,(B-T + R-L)*(1.0/1.41));
	
	// Write back perturbed values onto terrain
	t(L,Y) = YL;
	t(R,Y) = YR;
	t(X,T) = TX;
	t(X,B) = BX;
	t(X,Y) = YX;
	
	// Recurse over four corners
	plasma(t,L,X, T,Y);
	plasma(t,L,X, Y,B);
	plasma(t,X,R, T,Y);
	plasma(t,X,R, Y,B);
}




void display(void) /* Draw everything in our window */
{
	oglCheck("before display function");
	static double elapsed=0.0; /* last elapsed display time (seconds) */
	static float display_dt=0.0;
	float big=300; /* size of the ground */
	float floor=0.0; /* Z height of floor */
	glFinish(); double start_t=oglTimer();
	
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	
	static terrain t(plasma_size+1,plasma_size+1); /* initialized below */
	static GLuint srcTex=0;
	if (srcTex==0) 
	{ /* Set up the terrain */
		init(t);
		plasma(t,0,t.w-1, 0,t.h-1);
		
		glGenTextures(1,&srcTex);
		glBindTexture(GL_TEXTURE_2D,srcTex);
		glTexImage2D(GL_TEXTURE_2D,0,
			GL_RGBA8, t.w,t.h,
			0, GL_LUMINANCE,GL_FLOAT,&t(0,0));
		glTexParameteri(GL_TEXTURE_2D,
			GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	}
  
/* Main camera setup */
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	glViewport(0,0,glutGet(GLUT_WINDOW_WIDTH),glutGet(GLUT_WINDOW_HEIGHT));
	glClearColor(0.3,0.5,0.7,1.0); /* clear to a nice blue color */
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(70.0, /* degree FOV */
		glutGet(GLUT_WINDOW_WIDTH)/float(glutGet(GLUT_WINDOW_HEIGHT)),
		0.1,1000.0 /* Z range */
	);
	
	/* Physics and display timesteps: */
	oglCameraLookAt(0.5);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
// Draw in our ground plane texture
	glBindTexture(GL_TEXTURE_2D,srcTex);
	glGenerateMipmapEXT(GL_TEXTURE_2D); /* make mipmaps before rendering */
	glEnable(GL_TEXTURE_2D);
	
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,magfilter_mode); 
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,minfilter_mode);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, wrap_mode); 
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, wrap_mode);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT,texaniso?16:1);

	glDisable(GL_BLEND);
	glBegin(GL_QUADS);
	glColor4f(1,1,1,1); 
	glTexCoord2f(-big,-big); glVertex3f(-big,-big,floor);
	glTexCoord2f(+big,-big); glVertex3f(+big,-big,floor);
	glTexCoord2f(+big,+big); glVertex3f(+big,+big,floor);
	glTexCoord2f(-big,+big); glVertex3f(-big,+big,floor);
	glEnd();
	glDisable(GL_TEXTURE_2D);
	
	oglCameraAxes();
	
	glutSwapBuffers(); /* display everything we've rendered so far */
	glutPostRedisplay(); /* continual animation */
	
	// Print frames per second
	glFinish(); elapsed=oglTimer()-start_t; /* stop the stopwatch */
	static double last_print=0.0; /* Periodically print out the time */
	if (start_t>last_print+0.1) {
		display_dt=elapsed+0.001;
		printf("%.1f frames/sec, %.2f ms/frame\n",
			1.0/elapsed,1000.0*elapsed);
		last_print=start_t;
	}
	oglCheck("after display function");
}

int main(int argc,char **argv) {
	glutInit(&argc,argv);  /* set up GLUT */
	
	glutInitWindowPosition(0,60);
	glutInitWindowSize(1000,700);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
	int win=glutCreateWindow("Fractal Terrain"); /* make a window */
	glewInit(); /* only AFTER create window! */
	
	glutDisplayFunc(display); /* here's how to draw the window */
	oglCameraInit();
	camera=vec3(0.5,0.5,0.80);
	camera_orient.y=vec3(0,1,0);camera_orient.z=vec3(0,0,1);
	glutMainLoop(); /* start drawing */
}
